@@ -8,11 +8,11 @@ HTML::ToDocBook - Converts an XHTML file into DocBook.
 
 =head1 VERSION
 
-This describes version B<0.02> of HTML::ToDocBook.
+This describes version B<0.03> of HTML::ToDocBook.
 
 =cut
 
-our $VERSION = '0.02';
+our $VERSION = '0.03';
 
 =head1 SYNOPSIS
 
@@ -22,23 +22,25 @@ our $VERSION = '0.02';
 
     $obj->convert(infile=>$filename);
 
+    # convert HTML file
+    $obj->convert(infile=>$filename, html=>1);
+
 =head1 DESCRIPTION
 
-This module converts an XHTML file into DocBook format using
-both heuristics and XSLT processing.  This expects the input file
-to be correct XHTML -- there are other programs (such as html tidy)
-http://tidy.sourceforge.net/
-which can correct files for you, this does not do that.
-
-(Note that if you use HTML Tidy, don't forget to set 'enclose-block-text'
-or any unenclosed text will dissappear.)
+This module converts an XHTML file into DocBook format using both
+heuristics and XSLT processing.  By default, this expects the input file to
+be correct XHTML -- there are other programs such as html tidy
+(http://tidy.sourceforge.net/) which can correct files for you; this does
+not do that.
 
 Note also this is very simple; it doesn't deal with things like
 <div> or <span> which it has no way of guessing the meaning of.
+(For some, however, if they have class names which match DocBook tags,
+they will be turned into those tags)
 This does not merge multiple XHTML files into a single document,
 so this converts each XHTML file into a <chapter>, with each
-header being a section (sect1 to sect5).  The first header is used
-for both the chapter title and the first section title.
+header being a section (sect1 to sect5).  The <title> tag is used for
+the chapter title.
 
 There will likely to be validity errors, depending on how good the original
 HTML was.  There may be broken links, <xref> elements that should be <link>s,
@@ -56,6 +58,22 @@ use HTML::SimpleParse;
 =head1 METHODS
 
 =head2 new
+    
+    my $conv = HTML::ToDocBook->new();
+
+    my $conv = HTML::ToDocBook->new(stylesheet=>$stylesheet);
+
+Arguments:
+
+=over
+
+=item stylesheet
+
+A replacement XSLT stylesheet to use for conversions instead of the
+built-in one.  This can either be a file name or a string containing
+the entire stylesheet.
+
+=back
 
 =cut
 
@@ -70,38 +88,75 @@ sub new {
     $self->{_parser} = $parser;
     $self->{_xslt} = $xslt;
 
-    # build the parsed stylesheet from the DATA
+    if ($self->{stylesheet}
+	and -f $self->{stylesheet})
+    {
+	my $fn = abs_path($self->{stylesheet});
+	my $style_doc = $parser->parse_file($fn) 
+	    or die "Could not parse $fn XSLT file";
+	my $stylesheet = $xslt->parse_stylesheet($style_doc)
+	    or die "Could not parse $fn stylesheet";
+	$self->{_xslt_sheet} = $stylesheet;
+    }
+    elsif ($self->{stylesheet})
+    {
+	my $style_doc = $parser->parse_string($self->{stylesheet})
+	    or die "Could not parse string XSLT";
+	my $stylesheet = $xslt->parse_stylesheet($style_doc)
+	    or die "Could not parse stylesheet";
+	$self->{_xslt_sheet} = $stylesheet;
+    }
+    else
+    {
 
-    # This is stored in the DATA handle, after the __DATA__ at
-    # the end of this file; but because the scripts may not just
-    # create one instance of this object,
-    # we have to remember the position of the DATA handle
-    # and reset it after we've read from it, just in case
-    # we have to read from it again.
-    # This also means that we don't close it, either.  Hope that doesn't
-    # cause a problem...
+	# build the parsed stylesheet from the DATA
 
-    my $curpos = tell(DATA);    # remember the __DATA__ position
-    my $style_doc = $parser->parse_fh(\*DATA);
-    # reset the data handle to the start, just in case
-    seek(DATA, $curpos, 0);
+	# This is stored in the DATA handle, after the __DATA__ at
+	# the end of this file; but because the scripts may not just
+	# create one instance of this object,
+	# we have to remember the position of the DATA handle
+	# and reset it after we've read from it, just in case
+	# we have to read from it again.
+	# This also means that we don't close it, either.  Hope that doesn't
+	# cause a problem...
 
-    my $stylesheet = $xslt->parse_stylesheet($style_doc);
-    $self->{_xslt_sheet} = $stylesheet;
+	my $curpos = tell(DATA);    # remember the __DATA__ position
+	my $style_doc = $parser->parse_fh(\*DATA);
+	# reset the data handle to the start, just in case
+	seek(DATA, $curpos, 0);
+
+	my $stylesheet = $xslt->parse_stylesheet($style_doc);
+	$self->{_xslt_sheet} = $stylesheet;
+    }
 
     return ($self);
 } # new
 
 =head2 convert
 
-    $obj->convert(infile=>$filename);
+    $obj->convert(infile=>$filename,
+		html=>1);
+
+Arguments:
+
+=over
+
+=item infile
+
+The name of the file to convert.
+
+=item html
+
+Parse the input as HTML rather than XML.
+
+=back
 
 =cut
 
 sub convert {
     my $self = shift;
     my %args = (
-	parse_type=>'xml',
+	html=>0,
 	@_
     );
     my $filename = $args{infile};
@@ -132,7 +187,7 @@ sub convert {
 
     my $source = undef;
     my $result_str = '';
-    if ($args{parse_type} =~ /html/i)
+    if ($args{html})
     {
 	$source = $self->{_parser}->parse_html_string($file_str);
     }
@@ -209,8 +264,13 @@ sub insert_sections {
 	    if (@levels)
 	    {
 		my $prev_level = $levels[$#levels];
-		if ($prev_level == $header_level
-		    or $prev_level > $header_level)
+		while ($prev_level > $header_level)
+		{
+		    pop @levels;
+		    push @newhtml, "</div>\n";
+		    $prev_level = $levels[$#levels];
+		}
+		if ($prev_level == $header_level)
 		{
 		    pop @levels;
 		    push @newhtml, "</div>\n";
@@ -221,7 +281,7 @@ sub insert_sections {
 	    push @levels, $header_level;
 	}
 	elsif ($tok->{type} eq 'endtag'
-	    and $tok->{content} =~ /body/i)
+	    and $tok->{content} =~ /^\/body/i)
 	{
 	    # we need to close any remaining open section divs
 	    while (@levels)
@@ -331,14 +391,19 @@ __DATA__
 <xsl:param name="prefix">wb</xsl:param>
 <xsl:param name="graphics_location">images/</xsl:param>
 
+<!-- This needs to match elements with both the html: namespace
+and without, because files parsed as HTML rather than XHTML
+don't have the html: namespace in them, for whatever reason.
+-->
+
 <!-- Main block-level conversions -->
-<xsl:template match="html:html">
- <xsl:apply-templates select="html:body"/>
+<xsl:template match="html:html|html">
+ <xsl:apply-templates select="html:body|body"/>
 </xsl:template>
 
 <!-- This template converts each HTML file encountered into a DocBook
-     chapter.  For a title, it selects the first h1 element -->
-<xsl:template match="html:body">
+     chapter.  For a title, it selects the title, else the first h1 element -->
+<xsl:template match="html:body|body">
  <chapter>
   <xsl:if test="$filename != ''">
    <xsl:attribute name="id">
@@ -348,7 +413,9 @@ __DATA__
    </xsl:attribute>
   </xsl:if>
   <title>
-   <xsl:value-of select=".//html:h1[1]
+   <xsl:value-of select="/html:html/html:head/html:title
+			 |/html/head/title
+                         |.//html:h1[1]
                          |.//html:h2[1]
                          |.//html:h3[1]"/>
   </title>
@@ -359,32 +426,32 @@ __DATA__
 <!-- Sections
 These expect div class="sectN" tags
 -->
-<xsl:template match="html:div[@class='sect1']">
+<xsl:template match="html:div[@class='sect1']|div[@class='sect1']">
   <sect1>
     <xsl:apply-templates/>
   </sect1>
 </xsl:template>
-<xsl:template match="html:div[@class='sect2']">
+<xsl:template match="html:div[@class='sect2']|div[@class='sect2']">
   <sect2>
     <xsl:apply-templates/>
   </sect2>
 </xsl:template>
-<xsl:template match="html:div[@class='sect3']">
+<xsl:template match="html:div[@class='sect3']|div[@class='sect3']">
   <sect3>
     <xsl:apply-templates/>
   </sect3>
 </xsl:template>
-<xsl:template match="html:div[@class='sect4']">
+<xsl:template match="html:div[@class='sect4']|div[@class='sect4']">
   <sect4>
     <xsl:apply-templates/>
   </sect4>
 </xsl:template>
-<xsl:template match="html:div[@class='sect5']">
+<xsl:template match="html:div[@class='sect5']|div[@class='sect5']">
   <sect5>
     <xsl:apply-templates/>
   </sect5>
 </xsl:template>
-<xsl:template match="html:div[@class='section']">
+<xsl:template match="html:div[@class='section']|div[@class='section']">
   <section>
     <xsl:apply-templates/>
   </section>
@@ -396,7 +463,7 @@ above)
 It attempts to assign an ID to each sect by looking
 for a named anchor as a child of the header or as the immediate preceding
 or following sibling -->
-<xsl:template match="html:h1|html:h2|html:h3|html:h4|html:h5">
+<xsl:template match="html:h1|html:h2|html:h3|html:h4|html:h5|h1|h2|h3|h4|h5">
  <title>
   <xsl:choose>
    <xsl:when test="count(html:a/@name)">
@@ -404,24 +471,63 @@ or following sibling -->
      <xsl:value-of select="html:a/@name"/>
     </xsl:attribute>
    </xsl:when>
+   <xsl:when test="count(a/@name)">
+    <xsl:attribute name="id">
+     <xsl:value-of select="a/@name"/>
+    </xsl:attribute>
+   </xsl:when>
    <xsl:when test="preceding-sibling::* = preceding-sibling::html:a[@name != '']">
     <xsl:attribute name="id">
     <xsl:value-of select="concat($prefix,preceding-sibling::html:a[1]/@name)"/>
     </xsl:attribute>
    </xsl:when>
+   <xsl:when test="preceding-sibling::* = preceding-sibling::a[@name != '']">
+    <xsl:attribute name="id">
+    <xsl:value-of select="concat($prefix,preceding-sibling::a[1]/@name)"/>
+    </xsl:attribute>
+   </xsl:when>
    <xsl:when test="following-sibling::* = following-sibling::html:a[@name != '']">
     <xsl:attribute name="id">
     <xsl:value-of select="concat($prefix,following-sibling::html:a[1]/@name)"/>
     </xsl:attribute>
    </xsl:when>
+   <xsl:when test="following-sibling::* = following-sibling::a[@name != '']">
+    <xsl:attribute name="id">
+    <xsl:value-of select="concat($prefix,following-sibling::a[1]/@name)"/>
+    </xsl:attribute>
+   </xsl:when>
   </xsl:choose>
   <xsl:apply-templates/>
  </title>
 </xsl:template>
 
+<!-- These templates check class names to see if they are DocBook
+tag names.  This only does it for a few tags, though.
+-->
+<xsl:template match="html:span[@class='guilabel']|span[@class='guilabel']">
+<guilabel>
+  <xsl:apply-templates/>
+</guilabel>
+</xsl:template>
+<xsl:template match="html:span[@class='guibutton']|span[@class='guibutton']">
+<guibutton>
+  <xsl:apply-templates/>
+</guibutton>
+</xsl:template>
+<xsl:template match="html:span[@class='guimenu']|span[@class='guimenu']">
+<guimenuitem>
+  <xsl:apply-templates/>
+</guimenuitem>
+</xsl:template>
+<xsl:template match="html:span[@class='guimenuitem']|span[@class='guimenuitem']">
+<guimenuitem>
+  <xsl:apply-templates/>
+</guimenuitem>
+</xsl:template>
+
 <!-- These templates perform one-to-one conversions of HTML elements into
      DocBook elements -->
-<xsl:template match="html:p">
+<xsl:template match="html:p|p">
 <!-- if the paragraph has no text (perhaps only a child <img>), don't
      make it a para -->
  <xsl:choose>
@@ -435,14 +541,15 @@ or following sibling -->
   </xsl:otherwise>
  </xsl:choose>
 </xsl:template>
-<xsl:template match="html:pre">
+
+<xsl:template match="html:pre|pre">
  <programlisting>
   <xsl:apply-templates/>
  </programlisting>
 </xsl:template>
 
 <!-- Hyperlinks -->
-<xsl:template match="html:a[contains(@href,'http://')]" priority="1.5">
+<xsl:template match="html:a[contains(@href,'http://')]|a[contains(@href, 'http://')]" priority="1.5">
  <ulink>
   <xsl:attribute name="url">
    <xsl:value-of select="normalize-space(@href)"/>
@@ -451,7 +558,7 @@ or following sibling -->
  </ulink>
 </xsl:template>
 
-<xsl:template match="html:a[contains(@href,'#')]" priority="0.6">
+<xsl:template match="html:a[contains(@href,'#')]|a[contains(@href, '#')]" priority="0.6">
  <xref>
   <xsl:attribute name="linkend">
    <xsl:call-template name="make_id">
@@ -461,7 +568,7 @@ or following sibling -->
  </xref>
 </xsl:template>
 
-<xsl:template match="html:a[@href != '']">
+<xsl:template match="html:a[@href != '']|a[@href != '']">
  <xref>
   <xsl:attribute name="linkend">
    <xsl:value-of select="$prefix"/>
@@ -513,13 +620,17 @@ or following sibling -->
 
 <!-- Images -->
 <!-- Images and image maps -->
-<xsl:template match="html:img">
+<xsl:template match="html:img|img">
  <xsl:variable name="tag_name">
   <xsl:choose>
    <xsl:when test="boolean(parent::html:p) and
         boolean(normalize-space(parent::html:p/text()))">
     <xsl:text>inlinemediaobject</xsl:text>
    </xsl:when>
+   <xsl:when test="boolean(parent::p) and
+        boolean(normalize-space(parent::p/text()))">
+    <xsl:text>inlinemediaobject</xsl:text>
+   </xsl:when>
    <xsl:otherwise>mediaobject</xsl:otherwise>
   </xsl:choose>
  </xsl:variable>
@@ -560,7 +671,7 @@ or following sibling -->
  <xsl:value-of select="$graphics_location"/><xsl:value-of select="$name_only"/>
 </xsl:template>
 
-<xsl:template match="html:ul[count(*) = 0]">
+<xsl:template match="html:ul[count(*) = 0]|ul[count(*) = 0]">
  <xsl:message>Matched</xsl:message>
  <blockquote>
   <xsl:apply-templates/>
@@ -582,39 +693,42 @@ or following sibling -->
 </xsl:template>
 
 <!-- LIST ELEMENTS -->
-<xsl:template match="html:ul">
+<xsl:template match="html:ul|ul">
  <itemizedlist>
   <xsl:apply-templates/>
  </itemizedlist>
 </xsl:template>
 
-<xsl:template match="html:ol">
+<xsl:template match="html:ol|ol">
  <orderedlist>
   <xsl:apply-templates/>
  </orderedlist>
 </xsl:template>
 
 <!-- This template makes a DocBook variablelist out of an HTML definition list -->
-<xsl:template match="html:dl">
+<xsl:template match="html:dl|dl">
  <variablelist>
-  <xsl:for-each select="html:dt">
+  <xsl:for-each select="html:dt|dt">
    <varlistentry>
     <term>
      <xsl:apply-templates/>
     </term>
     <listitem>
-     <xsl:apply-templates select="following-sibling::html:dd[1]"/>
+     <xsl:apply-templates select="following-sibling::html:dd[1]|following-sibling::dd[1]"/>
     </listitem>
    </varlistentry>
   </xsl:for-each>
  </variablelist>
 </xsl:template>
 
-<xsl:template match="html:dd">
+<xsl:template match="html:dd|dd">
  <xsl:choose>
   <xsl:when test="boolean(html:p)">
    <xsl:apply-templates/>
   </xsl:when>
+  <xsl:when test="boolean(p)">
+   <xsl:apply-templates/>
+  </xsl:when>
   <xsl:otherwise>
    <para>
     <xsl:apply-templates/>
@@ -623,7 +737,7 @@ or following sibling -->
  </xsl:choose>
 </xsl:template>
 
-<xsl:template match="html:li">
+<xsl:template match="html:li|li">
  <listitem>
   <xsl:choose>
    <xsl:when test="count(html:p) = 0">
@@ -631,6 +745,11 @@ or following sibling -->
      <xsl:apply-templates/>
     </para>
    </xsl:when>
+   <xsl:when test="count(p) = 0">
+    <para>
+     <xsl:apply-templates/>
+    </para>
+   </xsl:when>
    <xsl:otherwise>
     <xsl:apply-templates/>
    </xsl:otherwise>
@@ -639,7 +758,7 @@ or following sibling -->
 </xsl:template>
 
 <xsl:template match="*">
- <xsl:message>No template for <xsl:value-of select="name()"/>
+ <xsl:message>No template for <xsl:value-of select="name()"/><xsl:text> </xsl:text><xsl:value-of select="@class"/>
  </xsl:message>
  <xsl:apply-templates/>
 </xsl:template>
@@ -651,26 +770,27 @@ or following sibling -->
 </xsl:template>
 
 <!-- inline formatting -->
-<xsl:template match="html:b|html:strong">
+<xsl:template match="html:b|html:strong|b|strong">
  <emphasis role="bold">
   <xsl:apply-templates/>
  </emphasis>
 </xsl:template>
-<xsl:template match="html:i|html:em">
+<xsl:template match="html:i|html:em|i|em">
  <emphasis>
   <xsl:apply-templates/>
  </emphasis>
 </xsl:template>
-<xsl:template match="html:u">
+<xsl:template match="html:u|u">
  <citetitle>
   <xsl:apply-templates/>
  </citetitle>
 </xsl:template>
 
 <!-- Ignored elements -->
-<xsl:template match="html:hr"/>
-<xsl:template match="html:br"/>
+<xsl:template match="html:hr|hr"/>
+<xsl:template match="html:br|br"/>
 <xsl:template match="html:p[normalize-space(.) = '' and count(*) = 0]"/>
+<xsl:template match="p[normalize-space(.) = '' and count(*) = 0]"/>
 
 <xsl:template match="text()">
  <xsl:choose>
@@ -680,14 +800,15 @@ or following sibling -->
 </xsl:template>
 
 <!-- Workbench Hacks -->
-<xsl:template match="html:div[contains(@style,'margin-left: 2em')]">
+<xsl:template match="html:div[contains(@style,'margin-left: 2em')]|div[contains(@style,'margin-left: 2em')]">
  <blockquote><para>
   <xsl:apply-templates/></para>
  </blockquote>
 </xsl:template>
 
 <xsl:template match="html:a[@href != ''
-                      and not(boolean(ancestor::html:p|ancestor::html:li))]"
+                      and not(boolean(ancestor::html:p|ancestor::html:li))]
+|a[@href != '' and not(boolean(ancestor::p|ancestor::li))]"
               priority="1">
  <para>
  <xref>
@@ -703,7 +824,9 @@ or following sibling -->
 </xsl:template>
 
 <xsl:template match="html:a[contains(@href,'#')
-                    and not(boolean(ancestor::html:p|ancestor::html:li))]"
+                    and not(boolean(ancestor::html:p|ancestor::html:li))]
+|a[contains(@href,'#')
+                    and not(boolean(ancestor::p|ancestor::li))]"
               priority="1.1">
  <para>
  <xref>
@@ -720,7 +843,7 @@ or following sibling -->
 
 <!-- ================================================================= -->
 <!-- Table conversion -->
-<xsl:template match="html:table">
+<xsl:template match="html:table|table">
  <xsl:variable name="column_count">
   <xsl:call-template name="count_columns">
    <xsl:with-param name="table" select="."/>
@@ -735,10 +858,10 @@ or following sibling -->
     <xsl:with-param name="count" select="$column_count"/>
    </xsl:call-template>
    <thead>
-    <xsl:apply-templates select="html:tr[1]"/>
+    <xsl:apply-templates select="html:tr[1]|*/tr[1]|tr[1]|*/tr[1]"/>
    </thead>
    <tbody>
-    <xsl:apply-templates select="html:tr[position() != 1]"/>
+    <xsl:apply-templates select="html:tr[position() != 1]|*/tr[position() != 1]|tr[position() != 1]|*/tr[position() != 1]"/>
    </tbody>
   </tgroup>
  </informaltable>
@@ -766,13 +889,13 @@ or following sibling -->
  </xsl:choose>
 </xsl:template>
 
-<xsl:template match="html:tr">
+<xsl:template match="html:tr|tr">
  <row>
   <xsl:apply-templates/>
  </row>
 </xsl:template>
 
-<xsl:template match="html:th|html:td">
+<xsl:template match="html:th|html:td|th|td">
  <xsl:variable name="position" select="count(preceding-sibling::*) + 1"/>
  <entry>
   <xsl:if test="@colspan &gt; 1">
@@ -792,13 +915,13 @@ or following sibling -->
  </entry>
 </xsl:template>
 
-<xsl:template match="html:td_null">
+<xsl:template match="html:td_null|td_null">
  <xsl:apply-templates/>
 </xsl:template>
 
 <xsl:template name="count_columns">
  <xsl:param name="table" select="."/>
- <xsl:param name="row" select="$table/html:tr[1]"/>
+ <xsl:param name="row" select="$table/html:tr[1]|$table/*/html:tr[1]|$table/tr[1]|$table/*/tr[1]"/>
  <xsl:param name="max" select="0"/>
  <xsl:choose>
   <xsl:when test="local-name($table) != 'table'">
@@ -811,7 +934,7 @@ or following sibling -->
    <!-- Count cells in the current row -->
    <xsl:variable name="current_count">
     <xsl:call-template name="count_cells">
-     <xsl:with-param name="cell" select="$row/html:td[1]|$row/html:th[1]"/>
+     <xsl:with-param name="cell" select="$row/html:td[1]|$row/html:th[1]|$row/td[1]|$row/th[1]"/>
     </xsl:call-template>
    </xsl:variable>
    <!-- Check for the maximum value of $current_count and $max -->
@@ -830,10 +953,13 @@ or following sibling -->
     <xsl:when test="count($row/following-sibling::html:tr) = 0">
      <xsl:value-of select="$new_max"/>
     </xsl:when>
+    <xsl:when test="count($row/following-sibling::tr) = 0">
+     <xsl:value-of select="$new_max"/>
+    </xsl:when>
     <xsl:otherwise>
      <xsl:call-template name="count_columns">
       <xsl:with-param name="table" select="$table"/>
-      <xsl:with-param name="row" select="$row/following-sibling::html:tr"/>
+      <xsl:with-param name="row" select="$row/following-sibling::html:tr|$row/following-sibling::tr"/>
       <xsl:with-param name="max" select="$new_max"/>
      </xsl:call-template>
     </xsl:otherwise>
